@@ -4,22 +4,23 @@ from __future__ import annotations
 
 import argparse
 import ast
-import importlib
+import importlib.util
 import itertools
 import logging
-import os
+import pprint
 import random
 import sys
-from collections import defaultdict
+from collections import defaultdict, OrderedDict
+from dataclasses import dataclass
 from pathlib import Path
 from typing import (
     DefaultDict,
+    Final,
     Iterable,
     List,
     Literal,
     Mapping,
     NamedTuple,
-    Optional,
     Sequence,
     Set,
     Tuple,
@@ -31,7 +32,7 @@ from graphviz import Digraph
 
 from py_import_cycles.dfs import depth_first_search
 from py_import_cycles.tarjan import strongly_connected_components as tarjan_scc
-from py_import_cycles.typing import Comparable
+from py_import_cycles.type_defs import Comparable
 
 logger = logging.getLogger(__name__)
 
@@ -59,7 +60,7 @@ logger = logging.getLogger(__name__)
 # Handle relative imports properly
 
 # TODO #7
-# Exclude specific folders like tests...
+# Exclude specific packages like tests...
 
 # TODO #8
 # more log levels: verbose: log nrealy every step
@@ -89,85 +90,134 @@ logger = logging.getLogger(__name__)
 #   '----------------------------------------------------------------------'
 
 
+class ModuleName:
+    """This class is inspired by pathlib.Path"""
+
+    def __init__(self, *parts: str | ModuleName) -> None:
+        self._parts: Final[tuple[str, ...]] = tuple(
+            entry
+            for part in parts
+            for entry in (part.parts if isinstance(part, ModuleName) else part.split("."))
+        )
+
+    def __str__(self) -> str:
+        return ".".join(self._parts)
+
+    def __hash__(self) -> int:
+        return hash(self._parts)
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, ModuleName):
+            return NotImplemented
+        return self._parts == other._parts
+
+    def __lt__(self, other: object) -> bool:
+        if not isinstance(other, ModuleName):
+            return NotImplemented
+        return self._parts < other._parts
+
+    def __gt__(self, other: object) -> bool:
+        if not isinstance(other, ModuleName):
+            return NotImplemented
+        return self._parts > other._parts
+
+    @property
+    def parts(self) -> Sequence[str]:
+        return self._parts
+
+    @property
+    def parent(self) -> ModuleName:
+        return ModuleName(*self._parts[:-1])
+
+    @property
+    def parents(self) -> Sequence[ModuleName]:
+        return [ModuleName(*self._parts[:idx]) for idx in range(1, len(self._parts))][::-1]
+
+    def joinname(self, *names: str | ModuleName) -> ModuleName:
+        return ModuleName(*self._parts, *names)
+
+
 class RegularPackage(NamedTuple):
     path: Path
-    folder: Path
-    name: str
+    name: ModuleName
 
 
 class NamespacePackage(NamedTuple):
     path: Path
-    name: str
+    name: ModuleName
 
 
 class PyModule(NamedTuple):
     path: Path
-    name: str
+    name: ModuleName
 
 
 Module = Union[RegularPackage, NamespacePackage, PyModule]
 
 
-def _make_module_from_name(
-    mapping: Mapping[str, str], project_path: Path, module_name: str
-) -> Optional[Module]:
-    parts = module_name.split(".")
-    for key, value in mapping.items():
-        if value in parts:
-            parts = [key] + parts
-            break
+@dataclass(frozen=True)
+class ModuleFactory:
+    _mapping: Mapping[str, str]
+    _project_path: Path
 
-    module_path = project_path.joinpath(Path(*parts))
+    def make_module_from_name(self, module_name: ModuleName) -> Module | None:
+        def _get_sanitized_module_name() -> ModuleName:
+            for key, value in self._mapping.items():
+                if value in module_name.parts:
+                    return ModuleName(key).joinname(module_name)
+            return module_name
 
-    if module_path.is_dir():
-        if (init_module_path := module_path / "__init__.py").exists():
-            return RegularPackage(
-                path=init_module_path,
-                folder=module_path,
+        module_path = self._project_path.joinpath(Path(*_get_sanitized_module_name().parts))
+
+        if module_path.is_dir():
+            if (init_module_path := module_path / "__init__.py").exists():
+                return RegularPackage(
+                    path=init_module_path,
+                    name=module_name.joinname("__init__"),
+                )
+
+            return NamespacePackage(
+                path=module_path,
                 name=module_name,
             )
 
-        return NamespacePackage(
-            path=module_path,
-            name=module_name,
-        )
+        if (py_module_path := module_path.with_suffix(".py")).exists():
+            return PyModule(
+                path=py_module_path,
+                name=module_name,
+            )
 
-    if (py_module_path := module_path.with_suffix(".py")).exists():
-        return PyModule(
-            path=py_module_path,
-            name=module_name,
-        )
+        return None
 
-    return None
+    def make_module_from_path(self, module_path: Path) -> Module | None:
+        def _get_sanitized_module_name() -> ModuleName:
+            parts = module_path.relative_to(self._project_path).with_suffix("").parts
+            for key, value in self._mapping.items():
+                if key == parts[0] and value in parts[1:]:
+                    return ModuleName(*parts[1:])
+            return ModuleName(*parts)
 
+        module_name = _get_sanitized_module_name()
 
-def _make_module_from_path(
-    mapping: Mapping[str, str], project_path: Path, module_path: Path
-) -> Module:
-    def _make_module_name_from_path(
-        mapping: Mapping[str, str], project_path: Path, module_path: Path
-    ) -> str:
-        parts = module_path.relative_to(project_path).with_suffix("").parts
+        if module_path.is_dir():
+            return NamespacePackage(
+                path=module_path,
+                name=module_name,
+            )
 
-        for key, value in mapping.items():
-            if key == parts[0] and value in parts[1:]:
-                parts = parts[1:]
-                break
+        if module_path.is_file() and module_path.suffix == ".py":
+            if module_path.stem == "__init__":
+                return RegularPackage(
+                    path=module_path,
+                    name=module_name,
+                )
 
-        return ".".join(parts)
+            return PyModule(
+                path=module_path,
+                name=module_name,
+            )
 
-    if module_path.stem == "__init__":
-        folder = Path(*module_path.parts[:-1])
-        return RegularPackage(
-            path=module_path,
-            folder=folder,
-            name=_make_module_name_from_path(mapping, project_path, folder),
-        )
-
-    return PyModule(
-        path=module_path,
-        name=_make_module_name_from_path(mapping, project_path, module_path),
-    )
+        return None
 
 
 # .
@@ -181,9 +231,14 @@ def _make_module_from_path(
 #   '----------------------------------------------------------------------'
 
 
-def iter_python_files(project_path: Path, folders: Sequence[str]) -> Iterable[Path]:
-    for f in folders:
-        yield from (p.resolve() for p in (project_path / f).glob("**/*.py"))
+def iter_python_files(project_path: Path, packages: Sequence[str]) -> Iterable[Path]:
+    if packages:
+        for pkg in packages:
+            yield from (p.resolve() for p in (project_path / pkg).glob("**/*.py"))
+        return
+
+    for p in project_path.glob("**/*.py"):
+        yield p.resolve()
 
 
 # .
@@ -218,94 +273,94 @@ class NodeVisitorImports(ast.NodeVisitor):
 class ImportStmtsParser:
     def __init__(
         self,
-        mapping: Mapping[str, str],
-        project_path: Path,
+        module_factory: ModuleFactory,
         base_module: Module,
         import_stmts: Sequence[ImportSTMT],
     ) -> None:
-        self._mapping = mapping
-        self._project_path = project_path
+        self._module_factory = module_factory
         self._base_module = base_module
         self._import_stmts = import_stmts
 
     def get_imports(self) -> Iterable[Module]:
         for import_stmt in self._import_stmts:
             if isinstance(import_stmt, ast.Import):
-                yield from self._get_imported_modules_from_aliases(import_stmt.names)
+                yield from self._get_modules_of_import_stmt(import_stmt)
 
             elif isinstance(import_stmt, ast.ImportFrom):
-                yield from self._get_imported_from_modules(import_stmt)
+                yield from self._get_modules_of_import_from_stmt(import_stmt)
 
-    def _get_imported_from_modules(self, import_from_stmt: ast.ImportFrom) -> Iterable[Module]:
-        if not import_from_stmt.module:
-            return
+    # -----ast.Import-----
 
-        if self._is_builtin_or_stdlib(import_from_stmt.module):
-            return
+    def _get_modules_of_import_stmt(self, import_stmt: ast.Import) -> Iterable[Module]:
+        for alias in import_stmt.names:
+            if imported_module := self._get_module(ModuleName(alias.name)):
+                yield imported_module
 
-        if import_from_stmt.level == 0:
-            module_name = import_from_stmt.module
-        else:
-            module_name = ".".join(
-                self._base_module.name.split(".")[: -import_from_stmt.level]
-                + import_from_stmt.module.split(".")
-            )
+    # -----ast.ImportFrom-----
 
-        module = _make_module_from_name(
-            self._mapping,
-            self._project_path,
-            module_name,
-        )
-
-        if module is None:
-            logger.debug(
-                "Unhandled import in %s: %s",
-                self._base_module.name,
-                ast.dump(import_from_stmt),
-            )
-            return
-
-        if isinstance(module, PyModule):
-            yield module
-
-        elif isinstance(module, (RegularPackage, NamespacePackage)):
-            # TODO correct handling of NamespacePackage here?
-            yield from self._get_imported_modules_from_aliases(
-                import_from_stmt.names,
-                from_name=module.name,
-            )
-
-    def _get_imported_modules_from_aliases(
-        self, aliases: Sequence[ast.alias], from_name: str = ""
+    def _get_modules_of_import_from_stmt(
+        self, import_from_stmt: ast.ImportFrom
     ) -> Iterable[Module]:
-        for alias in aliases:
-            if self._is_builtin_or_stdlib(alias.name):
-                continue
+        if not (module_name_prefix := self._get_name_prefix(import_from_stmt)):
+            return
 
-            module = _make_module_from_name(
-                self._mapping,
-                self._project_path,
-                ".".join([from_name, alias.name]) if from_name else alias.name,
+        if imported_module := self._get_module(module_name_prefix):
+            yield imported_module
+
+        for alias in import_from_stmt.names:
+            # Add packages/modules to above prefix:
+            # 1 -> ../a/b/c{.py,/}
+            # 2 -> ../BASE/c{.py,/}
+            # 3 -> ../BASE/a/b/c{.py,/}
+            # 4 -> ../BASE_PARENT/a/b/c{.py,/}
+            if this_imported_module := self._get_module(module_name_prefix.joinname(alias.name)):
+                yield this_imported_module
+
+    def _get_name_prefix(self, import_from_stmt: ast.ImportFrom) -> ModuleName:
+        # Handle the cases:
+        # 1 from a.b import c (module == "a.b", level == 0)
+        #   -> Python module/package: ../a/b{.py,/}
+        # 2 from . import c (module == None, level == 1)
+        #   -> Python module/package: ../BASE{.py,/}
+        # 3 from .a.b import c (module == "a.b", level == 1)
+        #   -> Python module/package: ../BASE/a/b{.py,/}
+        # 4 from ..a.b import c (module == "a.b", level == 2)
+        #   -> Python module/package: ../BASE_PARENT/a/b{.py,/}
+        if import_from_stmt.level == 0:
+            return ModuleName(import_from_stmt.module) if import_from_stmt.module else ModuleName()
+
+        parent = self._base_module.name.parents[import_from_stmt.level - 1]
+        return parent.joinname(import_from_stmt.module) if import_from_stmt.module else parent
+
+    # -----helper-----
+
+    def _get_module(self, module_name: ModuleName) -> None | Module:
+        if self._is_builtin_or_stdlib(module_name):
+            return None
+
+        module = self._module_factory.make_module_from_name(module_name)
+
+        if (
+            module is None
+            or module == self._base_module
+            or (
+                isinstance(module, RegularPackage)
+                and module.name.parent == self._base_module.name.parent
             )
+        ):
+            # Last if-part: do not add reg pkg, ie. __init__.py, of base module
+            return None
 
-            if module is None:
-                logger.debug(
-                    "Unhandled import in %s: %s",
-                    self._base_module.name,
-                    ast.dump(alias),
-                )
-                continue
-
-            yield module
+        return module
 
     @staticmethod
-    def _is_builtin_or_stdlib(module_name: str) -> bool:
-        if module_name in sys.builtin_module_names or module_name in sys.modules:
+    def _is_builtin_or_stdlib(module_name: ModuleName) -> bool:
+        if str(module_name) in sys.builtin_module_names or str(module_name) in sys.modules:
             # Avail in 3.10: or name in sys.stdlib_module_names
             return True
 
         try:
-            return importlib.util.find_spec(module_name) is not None  # type: ignore[attr-defined]
+            return importlib.util.find_spec(str(module_name)) is not None
         except ModuleNotFoundError:
             return False
 
@@ -315,14 +370,10 @@ class ImportsOfModule(NamedTuple):
     imports: Sequence[Module]
 
 
-def _visit_python_file(
-    mapping: Mapping[str, str],
-    project_path: Path,
-    path: Path,
-) -> Optional[ImportsOfModule]:
-    module = _make_module_from_path(mapping, project_path, path)
+def _visit_python_file(module_factory: ModuleFactory, path: Path) -> None | ImportsOfModule:
+    module = module_factory.make_module_from_path(path)
 
-    if isinstance(module, NamespacePackage):
+    if module is None or isinstance(module, NamespacePackage):
         return None
 
     try:
@@ -342,16 +393,15 @@ def _visit_python_file(
     visitor.visit(tree)
 
     parser = ImportStmtsParser(
-        mapping,
-        project_path,
+        module_factory,
         module,
         visitor.import_stmts,
     )
 
-    if imports := list(parser.get_imports()):
-        return ImportsOfModule(module, imports)
-
-    return None
+    return ImportsOfModule(
+        module,
+        list(OrderedDict.fromkeys(parser.get_imports())),
+    )
 
 
 # .
@@ -396,21 +446,21 @@ def _make_graph(
     with d.subgraph() as ds:
         for edge in edges:
             ds.node(
-                edge.from_module.name,
+                str(edge.from_module.name),
                 shape=_get_shape(edge.from_module),
             )
 
             ds.node(
-                edge.to_module.name,
+                str(edge.to_module.name),
                 shape=_get_shape(edge.to_module),
             )
 
             ds.attr("edge", color=edge.edge_color)
 
             if edge.title:
-                ds.edge(edge.from_module.name, edge.to_module.name, edge.title)
+                ds.edge(str(edge.from_module.name), str(edge.to_module.name), edge.title)
             else:
-                ds.edge(edge.from_module.name, edge.to_module.name)
+                ds.edge(str(edge.from_module.name), str(edge.to_module.name))
 
     return d.unflatten(stagger=50)
 
@@ -482,7 +532,7 @@ def _make_edges(
             return _make_dfs_import_edges(import_cycles)
         return _make_only_cycles_edges(import_cycles)
 
-    raise NotImplementedError("Unknown graph option: %s" % args.graph)
+    raise NotImplementedError(f"Unknown graph option: {args.graph}")
 
 
 def _make_all_edges(
@@ -535,9 +585,9 @@ def _make_dfs_import_edges(
         nn = int(255 * normalize(value, -bound, bound + 1))
         assert 0 <= nn < 256
         if value < 0:
-            color = "#%02x%02x%02x" % (0, 255 - nn, nn)
+            color = "#%02x%02x%02x" % (0, 255 - nn, nn)  # pylint: disable=consider-using-f-string
         else:
-            color = "#%02x%02x%02x" % (nn, 0, 255 - nn)
+            color = "#%02x%02x%02x" % (nn, 0, 255 - nn)  # pylint: disable=consider-using-f-string
         out.append(ImportEdge(str(value), edge[0], edge[1], color))
     return out
 
@@ -547,7 +597,7 @@ def _make_only_cycles_edges(
 ) -> Sequence[ImportEdge]:
     edges: Set[ImportEdge] = set()
     for nr, import_cycle in enumerate(import_cycles, start=1):
-        color = "#%02x%02x%02x" % (
+        color = "#%02x%02x%02x" % (  # pylint: disable=consider-using-f-string
             random.randint(50, 200),
             random.randint(50, 200),
             random.randint(50, 200),
@@ -557,7 +607,7 @@ def _make_only_cycles_edges(
         for next_module in import_cycle[1:]:
             edges.add(
                 ImportEdge(
-                    "%s (%s)" % (str(nr), len(import_cycle) - 1),
+                    f"{str(nr)} ({len(import_cycle) - 1})",
                     start_module,
                     next_module,
                     color,
@@ -615,13 +665,15 @@ def _parse_arguments() -> argparse.Namespace:
     parser.add_argument(
         "--project-path",
         required=True,
-        help="path to project",
+        help=(
+            "path to project. If no packages are given collect all Python"
+            " files from this directory."
+        ),
     )
     parser.add_argument(
-        "--folders",
+        "--packages",
         nargs="+",
-        required=True,
-        help="collect Python files from top-level folders",
+        help="collect Python files from top-level packages",
     )
     parser.add_argument(
         "--strategy",
@@ -630,12 +682,6 @@ def _parse_arguments() -> argparse.Namespace:
         help="path-based strong component algorithm",
     )
 
-    parser.add_argument(
-        "--store-cycles",
-        type=str,
-        help="Write cycles to file instead of printing on stderr",
-    )
-
     return parser.parse_args()
 
 
@@ -644,13 +690,16 @@ class OutputsFilepaths(NamedTuple):
     graph: Path
 
 
-def _get_outputs_filepaths(args: argparse.Namespace) -> OutputsFilepaths:
-    target_dir = Path(os.path.abspath(__file__)).parent.parent.joinpath("outputs")
+def _get_outputs_filepaths(project_path: Path, packages: Sequence[str]) -> OutputsFilepaths:
+    target_dir = Path.home() / Path(".local", "py_import_cycles", "outputs")
     target_dir.mkdir(parents=True, exist_ok=True)
-    filename = "%s-%s" % (
-        "-".join(Path(args.project_path).parts[1:]),
-        "-".join(sorted(args.folders)),
-    )
+
+    filename_parts = list(Path(project_path).parts[1:])
+    if packages:
+        filename_parts.extend(sorted(packages))
+
+    filename = "-".join(filename_parts)
+
     return OutputsFilepaths(
         log=(target_dir / filename).with_suffix(".log"),
         graph=(target_dir / filename).with_suffix(".gv"),
@@ -659,8 +708,11 @@ def _get_outputs_filepaths(args: argparse.Namespace) -> OutputsFilepaths:
 
 def _setup_logging(args: argparse.Namespace, outputs_filepaths: OutputsFilepaths) -> None:
     log_filepath = outputs_filepaths.log
+
+    log_filepath.unlink(missing_ok=True)
+
     if args.debug:
-        sys.stderr.write("Write log to %s\n" % log_filepath)
+        sys.stderr.write(f"Write log to {log_filepath}\n")
         log_level = logging.DEBUG
     else:
         log_level = logging.NOTSET
@@ -675,25 +727,24 @@ def _setup_logging(args: argparse.Namespace, outputs_filepaths: OutputsFilepaths
     logger.addHandler(handler)
 
 
-def _show_or_store_cycles(
-    args: argparse.Namespace,
-    import_cycles: Sequence[Tuple[Module, ...]],
-) -> None:
-    if not import_cycles:
-        return
+def _log_or_show_cycles(import_cycles: Sequence[Tuple[Module, ...]], verbose: bool) -> None:
+    sys.stderr.write(f"Found {len(import_cycles)} import cycles\n")
 
-    sys.stderr.write("Found %d import cycles\n" % len(import_cycles))
-
-    if not args.store_cycles:
-        if not args.verbose:
-            return
+    if logger.level == logging.DEBUG:
+        # Avoid execution of pprint.pformat call if not debug
+        logger.debug(
+            "Import cycles: %s",
+            pprint.pformat(
+                {
+                    nr: [str(p.name) for p in import_cycle]
+                    for nr, import_cycle in enumerate(import_cycles, start=1)
+                }
+            ),
+        )
 
+    if verbose:
         for nr, import_cycle in enumerate(import_cycles, start=1):
-            sys.stderr.write("  %d: %s\n" % (nr, [ic.name for ic in import_cycle]))
-        return
-
-    with Path(args.store_cycles).open("w") as f:
-        f.write("\n".join([str(ic) for ic in import_cycles]))
+            sys.stderr.write(f"  {nr}: {[str(ic.name) for ic in import_cycle]}\n")
 
 
 # .
@@ -702,27 +753,38 @@ def _show_or_store_cycles(
 def main() -> int:
     args = _parse_arguments()
 
-    outputs_filepaths = _get_outputs_filepaths(args)
+    project_path = Path(args.project_path)
+    packages = args.packages if args.packages else []
+    mapping = {} if args.map is None else dict([entry.split(":") for entry in args.map])
+
+    if not project_path.exists() or not project_path.is_dir():
+        sys.stderr.write(f"No such directory: {project_path}\n")
+        return 1
+
+    outputs_filepaths = _get_outputs_filepaths(project_path, packages)
 
     _setup_logging(args, outputs_filepaths)
 
-    project_path = Path(args.project_path)
-    if not project_path.exists() or not project_path.is_dir():
-        sys.stderr.write("No such directory: %s\n" % project_path)
-        return 1
-
-    mapping = {} if args.map is None else dict([entry.split(":") for entry in args.map])
-
     logger.info("Get Python files")
-    python_files = iter_python_files(project_path, args.folders)
+    python_files = iter_python_files(project_path, packages)
 
     logger.info("Visit Python files, get imports by module")
+    module_factory = ModuleFactory(mapping, project_path)
     imports_by_module = {
         visited.module: visited.imports
         for path in python_files
-        if (visited := _visit_python_file(mapping, project_path, path)) is not None
+        if (visited := _visit_python_file(module_factory, path)) is not None
     }
 
+    if logger.level == logging.DEBUG:
+        # Avoid execution of pprint.pformat call if not debug
+        logger.debug(
+            "Imports by module: %s",
+            pprint.pformat(
+                {str(ibm.name): [str(m.name) for m in ms] for ibm, ms in imports_by_module.items()}
+            ),
+        )
+
     logger.info("Detect import cycles with strategy %s", args.strategy)
     unsorted_cycles = detect_cycles(args.strategy, imports_by_module)
     return_code = bool(unsorted_cycles)
@@ -735,7 +797,7 @@ def main() -> int:
         ((cycle[-1],) + cycle) for cycle in sorted_cycles
     ]
 
-    _show_or_store_cycles(args, import_cycles)
+    _log_or_show_cycles(import_cycles, args.verbose)
 
     if args.graph == "no":
         return return_code
